@@ -7,7 +7,7 @@ use warnings;
 use Time::HiRes qw(time);
 use Time::Local;
 
-our $VERSION = 1.16;
+our $VERSION = 1.22;
 
 sub new {
 	my ($class, %args) = @_;
@@ -20,6 +20,8 @@ sub new {
 	}, $class;
 	$self->{lru}{prev} = $self->{lru}{next} = $self->{lru};
 	$self->{session_expire} = 60 unless $self->{session_expire};
+	$self->{session_idle} = 1 unless $self->{session_idle};
+	$self->{session_persist_user} = '' unless $self->{session_persist_user};
 	$self->{max_user_session} = 0 unless $self->{max_user_session};
 	return $self;
 }
@@ -50,6 +52,18 @@ sub bin_search {
         return (0, $lo);
 }
 
+sub _expire_find_session_idx {
+	my ($self, $session, $idx) = @_;
+
+	do {
+		if ($self->{expiring}[$idx]->{id} eq $session->{id}) {
+			return $idx;
+		}
+		last if (--$idx < 0);
+	} while ($self->{expiring}[$idx]->{expiretime} eq $session->{expiretime});
+	return;
+}
+
 sub expire_insert {
 	my ($self, $session) = @_;
 
@@ -58,6 +72,14 @@ sub expire_insert {
 	#
 	my ($found, $idx) = bin_search($self->{expiring}, sub { $_[0]->{expiretime} - $_[1]->{expiretime} }, $session);
 	if ($found) {
+		# Update expire entry if session exists.
+		# (there should be just one session per id)
+		#
+		my $sessionidx = $self->_expire_find_session_idx($session, $idx);
+		if (defined $sessionidx) {
+			$self->{expiring}[$sessionidx] = $session;
+			return;
+		}
 		splice(@{$self->{expiring}}, $idx+1, 0, $session);
 	} else {
 		splice(@{$self->{expiring}}, $idx, 0, $session);
@@ -72,13 +94,10 @@ sub expire_remove {
 	#
 	my ($found, $idx) = bin_search($self->{expiring}, sub { $_[0]->{expiretime} - $_[1]->{expiretime} }, $session);
 	if ($found) {
-		do {
-			if ($self->{expiring}[$idx]->{id} eq $session->{id}) {
-				splice(@{$self->{expiring}}, $idx, 1);
-				last;
-			}
-			last if (--$idx < 0);
-		} while ($self->{expiring}[$idx]->{expiretime} eq $session->{expiretime});
+		my $sessionidx = $self->_expire_find_session_idx($session, $idx);
+		if (defined $sessionidx) {
+			splice(@{$self->{expiring}}, $sessionidx, 1);
+		}
 	}
 }
 
@@ -110,7 +129,7 @@ sub list_del {
 
 sub session_put {
 	my ($self, $child) = @_;
-	my $runtime = sprintf "%.02f", time() - $child->{start};
+	my %runtime = (exists $child->{runtime}) ? (runtime => $child->{runtime}) : ();
 
 	return unless exists $child->{session};
 
@@ -131,15 +150,17 @@ sub session_put {
 		}
 		$self->{per_user}{$user}{$child->{session}{id}} = 1;
 	}
-	$self->{trace_cb}->('PUT', {pid => $child->{pid}, id => $child->{id}, session => $child->{session}{id}, runtime => $runtime}) if $self->{trace_cb};
+	$self->{trace_cb}->('PUT', {pid => $child->{pid}, id => $child->{id}, session => $child->{session}{id}, %runtime}) if $self->{trace_cb};
 	$self->{active}{$child->{session}{id}} = $child;
 	list_add($self->{lru}{prev}, $child);
+	delete $child->{runtime};
 	return 1;
 }
 
 sub session_get {
-	my ($self, $session_id, $msg_id) = @_;
+	my ($self, $session_id, $msg_id, $msg_vci) = @_;
 	my %id = (defined $msg_id) ? (id => $msg_id) : ();
+	my %vci = (defined $msg_vci) ? (vci => $msg_vci) : ();
 
 	if (exists $self->{active}{$session_id}) {
 		my $child = delete $self->{active}{$session_id};
@@ -149,16 +170,50 @@ sub session_get {
 			my $user = $child->{session}{user};
 			if (exists $self->{per_user}{$user}) {
 				delete $self->{per_user}{$user}{$child->{session}{id}};
+				if (scalar keys %{$self->{per_user}{$user}} == 0) {
+					delete $self->{per_user}{$user};
+				}
 			}
 		}
 
 		my $stoptime = sprintf "%.02f", time() - $child->{start};
-		$self->{trace_cb}->('GET', {pid => $child->{pid}, %id, session => $session_id, stoptime => $stoptime}) if $self->{trace_cb};
+		$self->{trace_cb}->('GET', {pid => $child->{pid}, %id, %vci, session => $session_id, stoptime => $stoptime}) if $self->{trace_cb};
 		return $child;
 	}
 	return;
 }
 
+sub session_get_per_user {
+	my ($self, $user, $msg_id, $msg_vci) = @_;
+
+	if ($self->{max_user_session}) {
+		foreach my $session_id (keys %{$self->{per_user}{$user}}) {
+			return $self->session_get($session_id, $msg_id, $msg_vci);
+		}
+	}
+	return;
+}
+
+sub session_get_per_user_idle {
+	my ($self, $child) = @_;
+
+	return unless exists $child->{session};
+
+	if ($self->{max_user_session} && exists $child->{session}{user}) {
+		my $user = $child->{session}{user};
+		foreach my $session_id (keys %{$self->{per_user}{$user}}) {
+			if (exists $self->{active}{$session_id}) {
+				my $other_child = $self->{active}{$session_id};
+				my $idle = time() - $other_child->{start};
+				if ($idle >= $self->{session_idle}) {
+					return $self->session_get($session_id);
+				}
+			}
+		}
+	}
+	return;
+}
+
 sub parse_isotime {
 	my ($isotime) = @_;
 	my ($yy,$mm,$dd,$h,$m,$s,$msec) = $isotime =~
@@ -208,14 +263,14 @@ sub expired_dequeue {
 
 	# Use sorted expire list to expire sessions.
 	#
-	if (scalar @{$self->{expiring}}) {
+	while (scalar @{$self->{expiring}}) {
 		my $session = $self->{expiring}[0];
 		my $diff = $session->{expiretime} - time();
 		return if ($diff >= 0);
 
 		$session = shift @{$self->{expiring}};
 		my $child = $self->session_get($session->{id});
-		return $child;
+		return $child if $child;
 	}
 	return;
 }
